@@ -1,13 +1,19 @@
+import csv
 import json
 import logging
 import os
+from datetime import datetime
+from http import HTTPStatus
 from pathlib import Path
-from typing import List, Text, Optional
+from typing import List, Text, Optional, Tuple
 
 import requests
-import semantic_version
 from notebook.base.handlers import IPythonHandler
 from notebook.utils import url_path_join
+from pkg_resources import parse_version
+
+# noinspection PyProtectedMember
+from pkg_resources._vendor.packaging.version import Version
 
 import jupyter_cogram
 
@@ -31,10 +37,56 @@ logger.debug(f"PyPI URL: {package_pypi_url}")
 
 config_location = Path().home() / ".ipython/nbextensions/jupyter-cogram"
 token_file_name = "cogram_auth_token"
+log_file_name = "cogram_access_log"
+suggestions_timeout = float(os.environ.get("SUGGESTIONS_TIMEOUT", "10.0"))
 
 LATEST_PYPI_VERSION: Optional[Text] = None
 
 
+def read_csv(
+    p: Path = config_location / log_file_name,
+) -> Optional[List[Tuple[Text, Text]]]:
+    if not p.is_file():
+        logger.debug(f"'{p}' is not a file. Cannot read CSV.")
+        return None
+
+    out = []
+    with p.open() as f:
+        reader = csv.reader(f)
+        for row in reader:
+            k, v = row
+            out.append((k, v))
+
+    logger.debug(f"Successfully read CSV with {len(out)} entries.")
+    return out
+
+
+def create_log_file() -> None:
+    p = config_location.absolute() / log_file_name
+    if not p.is_file():
+        logger.info(f"Touching file '{p}'")
+        p.touch()
+
+
+def append_to_log_file() -> None:
+    line = f"{datetime.now().isoformat()},{jupyter_cogram.__version__}"
+
+    logger.info(f"Appending to log file '{line}'")
+    create_log_file()
+
+    p = config_location.absolute() / log_file_name
+
+    content = p.read_text()
+
+    with p.open("a") as f:
+        if not content:
+            to_write = line
+        else:
+            to_write = f"\n{line}"
+
+        f.write(to_write)
+
+
 def save_token(token: Text, loc: Path = config_location) -> None:
     loc = loc.absolute()
 
@@ -67,12 +119,12 @@ def post_auth_token(token: Text) -> requests.Response:
 
 
 def is_installation_up_to_date(
-        pypi_url: Text = package_pypi_url,
+    pypi_url: Text = package_pypi_url,
 ) -> bool:
     logger.info("Checking if local installation is up to date.")
 
     try:
-        res = requests.get(pypi_url, timeout=0.5)
+        res = requests.get(pypi_url, timeout=2)
     except Exception as e:
         logger.exception(e)
         return True
@@ -89,9 +141,9 @@ def is_installation_up_to_date(
         return True
 
     try:
-        is_up_to_date = semantic_version.Version(
-            jupyter_cogram.__version__
-        ) >= semantic_version.Version(LATEST_PYPI_VERSION)
+        is_up_to_date = parse_version(jupyter_cogram.__version__) >= parse_version(
+            LATEST_PYPI_VERSION
+        )
         logger.info(
             f"Have package version {jupyter_cogram.__version__}. "
             f"Latest PyPi version is {LATEST_PYPI_VERSION}. Installed version is up to "
@@ -104,44 +156,52 @@ def is_installation_up_to_date(
 
 
 def fetch_suggestion_code(
-        queries,
-        cell_contents: Optional[List[Text]] = None,
-        kernel_id: Optional[Text] = None,
-        session_id: Optional[Text] = None,
-        auth_token: Optional[Text] = None,
+    queries,
+    cell_contents: Optional[List[Text]] = None,
+    kernel_id: Optional[Text] = None,
+    session_id: Optional[Text] = None,
+    auth_token: Optional[Text] = None,
 ):
     url = f"{backend_url}/{completions_enpdoint}"
-    res = requests.post(
-        url,
-        json={
-            "queries": queries,
-            "cell_contents": cell_contents or [],
-            "kernel_id": kernel_id,
-            "session_id": session_id,
-            "auth_token": auth_token,
-        },
-    )
 
-    if res.status_code >= 400:
-        if res.json() and res.json().get("error") is not None:
-            msg = res.json()["error"]
-        else:
-            msg = res.text
+    try:
+        res = requests.post(
+            url,
+            json={
+                "queries": queries,
+                "cell_contents": cell_contents or [],
+                "kernel_id": kernel_id,
+                "session_id": session_id,
+                "auth_token": auth_token,
+            },
+            # timeout=suggestions_timeout,
+        )
 
+        if res.status_code >= 400:
+            if res.json() and res.json().get("error") is not None:
+                msg = res.json()["error"]
+            else:
+                msg = res.text
+
+            error_dict = {
+                "status": "error",
+                "error_msg": msg,
+                "status_code": res.status_code,
+            }
+
+            raise ValueError(json.dumps(error_dict))
+
+        return res.json()
+    except requests.exceptions.Timeout:
+        logger.error("Have Timeout")
         error_dict = {
             "status": "error",
-            "error_msg": msg,
-            "status_code": res.status_code,
+            "error_msg": "Request timed out",
+            "status_code": HTTPStatus.REQUEST_TIMEOUT,
         }
-
+        logger.debug(f"Submitting error dict {json.dumps(error_dict)}")
         raise ValueError(json.dumps(error_dict))
 
-    return res.json()
-
-
-def upgrade():
-    pass
-
 
 class JupyterCogramHandler(IPythonHandler):
     def __init__(self, application, request, **kwargs):
@@ -163,7 +223,10 @@ class JupyterCogramHandler(IPythonHandler):
 
             logger.debug(f"Have suggestions\n{json.dumps(suggestions, indent=2)}")
             if "choices" in suggestions and suggestions["choices"]:
-                message = [c.get("text", "") for c in suggestions["choices"]]
+                message = [
+                    {"suggestion": c.get("text", ""), "imports": c.get("imports")}
+                    for c in suggestions["choices"]
+                ]
             else:
                 message = None
 
@@ -184,7 +247,7 @@ class JupyterCogramHandler(IPythonHandler):
 
         response = {"status": status, "message": message}
         self.set_status(status_code)
-        # logger.debug(f"Returning response dict {json.dumps(response, indent=2)}")
+        logger.debug(f"Returning response dict {json.dumps(response, indent=2)}")
         self.finish(json.dumps(response))
 
 
@@ -299,9 +362,9 @@ class JupyterUpgradeCogramHandler(IPythonHandler):
         new_version = jupyter_cogram.__version__
         logger.debug(f"After upgrade, have jupyter-cogram version {new_version}")
 
-        if status_code == 0 and semantic_version.Version(
-                new_version
-        ) >= semantic_version.Version(previous_version):
+        if status_code == 0 and parse_version(new_version) >= parse_version(
+            previous_version
+        ):
             status = 200
             response = {
                 "status": "success",
@@ -313,8 +376,8 @@ class JupyterUpgradeCogramHandler(IPythonHandler):
             response = {
                 "status": "error",
                 "message": "Could not upgrade library. Please check your "
-                           "Jupyter Notebook server logs. If you keep encountering "
-                           "the same problem, please reach out at support@cogram.ai",
+                "Jupyter Notebook server logs. If you keep encountering "
+                "the same problem, please reach out at support@cogram.ai",
                 "new_version": LATEST_PYPI_VERSION,
             }
 
@@ -323,6 +386,64 @@ class JupyterUpgradeCogramHandler(IPythonHandler):
         self.finish(json.dumps(response))
 
 
+def get_new_version_message(
+    previous_version: Version,
+) -> Optional[Text]:
+    if previous_version.minor == 7:
+        return (
+            f"Welcome to your new Cogram version 👋 This version improves how you "
+            f"can write code with Cogram: You can now generate code by writing a "
+            f"Python "
+            f"comment and hitting the Tab key, or by beginning a line of code and "
+            f"completing with the Tab key."
+        )
+
+
+class JupyterCogramLaunchHandler(IPythonHandler):
+    def __init__(self, application, request, **kwargs):
+        super(JupyterCogramLaunchHandler, self).__init__(application, request, **kwargs)
+
+    def get(self):
+        logger.debug(f"Receiving launch query.")
+
+        log_content = read_csv()
+        is_first_launch = log_content is None
+
+        previous_version = log_content[-1][1] if log_content else None
+        is_new_version = (
+            parse_version(jupyter_cogram.__version__) > parse_version(previous_version)
+            if previous_version
+            else False
+        )
+
+        msg = None
+        if is_first_launch:
+            status = "first_launch"
+            msg = (
+                "Welcome to Cogram 👋 Get started by writing comments or code. You "
+                "can trigger completions with the Tab key."
+            )
+        elif is_new_version:
+            status = "new_version"
+            msg = get_new_version_message(parse_version(previous_version))
+        else:
+            status = "ok"
+
+        append_to_log_file()
+
+        self.set_status(200)
+        response = {
+            "status": status,
+            "version": jupyter_cogram.__version__,
+            "previous_version": previous_version,
+            "msg": msg,
+        }
+
+        logger.debug(f"Returning response dict {json.dumps(response, indent=2)}")
+
+        self.finish(json.dumps(response))
+
+
 def load_jupyter_server_extension(nb_server_app):
     """
     Called when the extension is loaded.
@@ -338,11 +459,12 @@ def load_jupyter_server_extension(nb_server_app):
         [
             (url_path_join(web_app.settings["base_url"], uri), handler)
             for uri, handler in [
-            ("/cogram", JupyterCogramHandler),
-            ("/token", JupyterCogramTokenHandler),
-            ("/checkVersion", JupyterCogramVersionHandler),
-            ("/upgrade", JupyterUpgradeCogramHandler),
-        ]
+                ("/cogram", JupyterCogramHandler),
+                ("/token", JupyterCogramTokenHandler),
+                ("/checkVersion", JupyterCogramVersionHandler),
+                ("/upgrade", JupyterUpgradeCogramHandler),
+                ("/launch", JupyterCogramLaunchHandler),
+            ]
         ],
     )
     logger.debug("loaded_jupyter_server_extension: jupyter-cogram")
